/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


function initBlacklistPane() {
  var subscriptLoader =  Cc["@mozilla.org/moz/jssubscript-loader;1"]
                           .getService(Ci.mozIJSSubScriptLoader);
  subscriptLoader.loadSubScript("chrome://global/content/nsDragAndDrop.js");
  subscriptLoader.loadSubScript("chrome://global/content/nsTransferable.js");

  let blacklist = Cc["@adblockfilters.mozdev.org/service;1"].getService()
                    .wrappedJSObject.blackWhiteList.concat([]);
  if(window.opener && opener.location.href == "chrome://adblockplus/content/settings.xul")
    blacklist = opener.gAdblockFilterExport.blackWhiteList;
  let tree = document.getElementById("blacklist");
  gBlackListView = tree.view = new BlackListView(blacklist);

  window.addEventListener("dialogaccept", blacklistDialogAccept, false);
  document.addEventListener("keypress", blacklistKeyPress, true);
  tree.addEventListener("dblclick", blacklistDoubleClick, false);
}
function blacklistDialogAccept() {
  let service = Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject;
  service.blackWhiteList = gBlackListView.blacklist;
  if(window.opener && opener.location == "chrome://adblockplus/content/settings.xul")
    opener.gAdblockFilterExport.blackWhiteList = gBlackListView.blacklist;
}
function blacklistKeyPress(event) {
  if(event.target.id != "blacklist")
    return;
  if([event.DOM_VK_ENTER, event.DOM_VK_RETURN].indexOf(event.keyCode) != -1) {
    gBlackListView._startOrStopEdit();
    event.stopPropagation();
    event.preventDefault();
  } else if(event.keyCode == event.DOM_VK_ESCAPE && event.target.editingColumn) {
    var row = event.target.editingRow, col = event.target.editingColumn;
    if(!gBlackListView.getCellText(row, col))
      gBlackListView.setCellText(row, col);
    gBlackListView._stopEdit(true);
    event.stopPropagation();
    event.preventDefault();
  }
}
function blacklistDoubleClick(event) {
  if(!gBlackListView._editableTree)
    gBlackListView._startEdit();
}

function focusTree() {
  let elem = document.getElementById('blacklist');
  if(elem.editingColumn)
    elem = elem.inputField;
  elem.focus();
}


var gBlackListView = null;
function BlackListView(blacklist) {
  this.blacklist = blacklist.concat([]);
}
BlackListView.prototype = {
  blacklist: [],
  get rowCount() {
    return this.blacklist.length;
  },
  get _lastIndex() {
    return this.rowCount - 1;
  },
  getCellText: function(row, column) {
    if(column.id == "filter") {
      return this.blacklist[row];
    }
    return null;
  },
  setCellText: function(row, column, text) {
    if(column.id == "filter") {
      var blacklist = this.blacklist;
      if(!text) {
        window.setTimeout(function(This) {
          This._remove(row);
        }, 0, this);
      } else if(blacklist.indexOf(text) != -1 && (blacklist.indexOf(text) != row ||
                blacklist.lastIndexOf(text) != row)) {
        window.setTimeout(function(This) {
          This._remove(row);
          This._select(blacklist.indexOf(text), true);
        }, 0, this);
      } else {
        blacklist[row] = text;
        this._tree.invalidateRow(row);
      }
    } else {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
  },
  _tree: null,
  _element: null,
  _editableTree: false,
  setTree: function(tree) {
    this._tree = tree;
    this._element = tree ? tree.element : null;
    this._editableTree = tree ? !!tree.element.startEditing : false;
  },

  _add: function() {
    var edit = this._element;
    if(edit.editingColumn) return;
    var idx = this._selectedIndex;
    if(idx == -1) idx = this._lastIndex;
    this.blacklist.splice(idx + 1, 0, "");
    this._tree.rowCountChanged(idx + 1, 1);
    this._select(idx + 1, true);
    this._startEdit();
  },

  get _selectedIndex() {
    return this.selection.currentIndex;
  },
  get _selectedIndexes() {
    var seln = this.selection;
    var indexes = [];
    for(var i = 0; i < this.rowCount; i++) {
      if(seln.isSelected(i))
        indexes.push(i);
    }
    return indexes.reverse();
  },
  _select: function (/*index0, index1, ...,  override*/) {
    var seln = this.selection;
    var override = arguments[arguments.length - 1];
    if(override) {
      seln.clearSelection();
    }

    for(var i = 0; i < arguments.length - 1; ++i) {
      if(arguments[i] < 0 || arguments[i] > this.rowCount) continue;
      if(!seln.isSelected(arguments[i]))
        seln.toggleSelect(arguments[i]);
      this._tree.ensureRowIsVisible(arguments[i]);
    }
  },
  selectionChanged: function() {
    var idx = this._selectedIndexes;
    var disable = (!idx.length);
    document.getElementById("cmd_edit").setAttribute("disabled", disable);
    document.getElementById("cmd_up").setAttribute("disabled", disable ||
                                                          idx.indexOf(0) != -1);
    document.getElementById("cmd_down").setAttribute("disabled", disable ||
                                            idx.indexOf(this._lastIndex) != -1);
    document.getElementById("cmd_remove").setAttribute("disabled", disable);
  },

  _bump: function(dir) {
    var selected = this._selectedIndexes;
    if(dir == -1)
      selected = selected.reverse();
    for(var i = 0; i < selected.length; i++) {
      var idx1 = Math.min(selected[i], selected[i] + dir);
      var idx2 = Math.max(selected[i], selected[i] + dir);
      this._tree.ensureRowIsVisible(selected[i]);
      this.blacklist.splice(idx1, 2, this.blacklist[idx2], this.blacklist[idx1]);
      this._tree.invalidateRow(idx1);
      this._tree.invalidateRow(idx2);
      selected[i] = selected[i] + dir;
    }
    selected.push(true);
    this._select.apply(this, selected);
  },
  _remove: function(row) {
    if(row !== undefined) {
      var selectedIndex = row;
      var selected = [row];
    } else {
      var selectedIndex = this._selectedIndex;
      var selected = this._selectedIndexes.reverse();
    }
    for(var i = 0; i < selected.length; i++) {
      this._tree.ensureRowIsVisible(selected[i]);
      this.blacklist.splice(selected[i], 1);
      this._tree.rowCountChanged(selected[i], -1);
      this._tree.invalidate();
    }
    this._select(Math.max(selectedIndex - 1, 0), true);
  },

  _promptNewVal: function(row, col) {
    var oldVal = this.getCellText(row, col);
    var value = { value: oldVal };
    var strings = "chrome://adblockfilters/locale/settings.properties";
    strings = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)
                .createBundle(strings);
    var strPrefix = "blacklist.edit.or.new.";
    var title = strings.formatStringFromName(strPrefix + "title", [oldVal], 1);
    var content = strings.formatStringFromName(strPrefix + "content", [oldVal], 1);
    var prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                    .getService(Ci.nsIPromptService);
    var abort = !prompts.prompt(window, title, content, value, null, {});
    if(abort)
      value.value = oldVal;
    this.setCellText(row, col, value.value);
    this._tree.ensureRowIsVisible(row);
  },
  _startEdit: function() {
    var column = this._tree.columns.getPrimaryColumn();
    if(this._editableTree)
      this._element.startEditing(this.selection.currentIndex, column);
    else
      this._promptNewVal(this.selection.currentIndex, column);
  },
  _stopEdit: function(abort) {
    if(this._editableTree)
      this._element.stopEditing(!abort);
  },
  _startOrStopEdit: function(abortStop) {
    if(!this._element.editingColumn)
      this._startEdit();
    else
      this._stopEdit(abortStop);
  },

  _startDrag: function(event) {
    if(event.target.nodeName == "treechildren") {
      nsDragAndDrop.startDrag(event, gBlacklistDragObserver);
    }
  },
  _getDragSourceItems: function() {
    var dragService = Cc["@mozilla.org/widget/dragservice;1"]
                        .getService(Ci.nsIDragService);
    var session = dragService.getCurrentSession();
    var data = Cc["@mozilla.org/widget/transferable;1"]
                 .createInstance(Ci.nsITransferable);
    data.addDataFlavor("text/unicode");
    session.getData(data, 0);
    var str = {}, len = {};
    data.getTransferData("text/unicode", str, len);
    var dragged = str.value.QueryInterface(Ci.nsISupportsString).data.substring(0, len.value);
    dragged = dragged.split("\n").filter(function(elem) { return elem; });
    return dragged;
  },
  canDrop: function(index, orientation) {
    var dragged = this._getDragSourceItems();
    return dragged.indexOf(this.blacklist[index]) == -1 &&
           dragged.indexOf(this.blacklist[index + orientation]) == -1;
  },
  drop: function(dropIndex, orientation) {
    if(dropIndex == -1)
      dropIndex = this._lastIndex, orientation = Ci.nsITreeView.DROP_AFTER;

    if(orientation == Ci.nsITreeView.DROP_AFTER) {
      dropIndex = dropIndex + 1;
    } else if(orientation != Ci.nsITreeView.DROP_BEFORE) {
      return
    }
    if(dropIndex < this.rowCount)
      var dropFilter = this.blacklist[dropIndex];
    else
      var dropFilter = null;

    var filters = this._getDragSourceItems();
    var list = [];
    for(var i = 0; i < this.rowCount; i++) {
      if(filters.indexOf(this.blacklist[i]) == -1) { // remains unchanged
        list.push(this.blacklist[i]);
      } else { // moved
        this._tree.rowCountChanged(i, -1);
      }
    }
    if(dropFilter == null)
      dropIndex = list.length;
    else
      dropIndex = list.indexOf(dropFilter);
    list.splice.apply(list, [dropIndex, 0].concat(filters)); // (re-)add filters
    this.blacklist = list;

    this._tree.rowCountChanged(dropIndex, filters.length);
    var newIndizes = [];
    for(var i = 0; i < filters.length; i++) {
      newIndizes.push(dropIndex + i);
    }
    newIndizes.push(true);
    this._select.apply(this, newIndizes);
  },

  isContainer: function(row){ return false; },
  isSeparator: function(row){ return false; },
  isSorted: function(){ return false; },
  getLevel: function(row){ return 0; },
  getImageSrc: function(row,col){ return null; },
  getRowProperties: function(row,props){},
  getCellProperties: function(row,col,props){},
  getColumnProperties: function(col, properties){ },
  cycleCell: function(row, col) {
    if(col.id == "remove") {
      this._remove(row);
    }
  },
  cycleHeader: function() {},
  getParentIndex: function() { return -1; },
  hasNextSibling: function(rowIndex) { return rowIndex < this.rowCount; },
  isEditable: function(row, col) {
    return (col.id == "filter");
  },
  toggleOpenState: function() {},
  getCellValue: function(row, col) { },
  performAction: function(action) {
    if(action == "delete")
      this._remove();
  },
  performActionOnRow: function(action, row) {
    if(action == "delete")
      this._remove(row);
  }
};

var gBlacklistDragObserver = {
  onDragStart: function (aEvent, aXferData, aDragAction) {
    var selected = gBlackListView._selectedIndexes;
    if(!selected.length)
      return;

    for(var i = 0; i < selected.length; i++) {
      selected[i] = gBlackListView.blacklist[selected[i]];
    }

    var str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
    str.data = selected.join("\n");
    aXferData.data = new TransferData();
    aXferData.data.addDataForFlavour("text/unicode", str, str.data.length * 2);

    aDragAction.action = Ci.nsIDragService.DRAGDROP_ACTION_MOVE;
  },
  onDrop: function (aEvent, aXferData, aDragSession) { },
  onDragExit: function (aEvent, aDragSession) { },
  onDragOver: function (aEvent) {
    if(aEvent.target.nodeName == "treechildren") {
      var row = {}, col = {}, part = {};
      aEvent.target.parentNode.treeBoxObject
                     .getCellAt(aEvent.clientX, aEvent.clientY, row, col, part);
      [row, col, part] = [row.value, col.value, part.value];
      if(row == -1) {
        var dragService = Cc["@mozilla.org/widget/dragservice;1"]
                            .getService(Ci.nsIDragService);
        var aDragSession = dragService.getCurrentSession();
        aDragSession.canDrop = true;
      }
    }
  },
  getSupportedFlavours: function() {
    return {
      "text/unicode": { iid: "nsISupportsString", width: 2 }
    };
  }
};
